#!/usr/bin/env python

"""
doctest runner for toolbox
"""

import doctest
import json
import os
import shutil
import sys
from cgi import escape
from optparse import OptionParser
from paste.fixture import TestApp
from time import time
from toolbox.dispatcher import Dispatcher

# global
directory = os.path.dirname(os.path.abspath(__file__))


class ToolboxTestApp(TestApp):
    """WSGI app wrapper for testing JSON responses"""

    def __init__(self, **kw):
        dispatcher_args = dict(model_type='memory_cache', fields=('usage', 'author', 'type', 'language', 'dependencies'))
        dispatcher_args.update(kw)
        app = Dispatcher(**dispatcher_args)
        TestApp.__init__(self, app)

    def get(self, url='/', **kwargs):
        kwargs.setdefault('params', {})['format'] = 'json'
        response = TestApp.get(self, url, **kwargs)
        return json.loads(response.body)

    def new(self, **kwargs):
        kwargs['form-render-date'] = str(time())
        return self.post('/new', params=kwargs)

    def cleanup(self):
        pass


class FileCacheTestApp(ToolboxTestApp):
    """test the MemoryCache file-backed backend"""

    def __init__(self):
        self.json_dir = os.path.join(directory, 'test_json')
        shutil.rmtree(self.json_dir, ignore_errors=True)
        os.makedirs(self.json_dir)
        ToolboxTestApp.__init__(self, model_type='file_cache', directory=self.json_dir)

    def cleanup(self):
        shutil.rmtree(self.json_dir, ignore_errors=True)


class CouchTestApp(ToolboxTestApp):
    """test the MemoryCache file-backed backend"""

    def __init__(self):
        ToolboxTestApp.__init__(self, model_type='couch', dbname='test_json')
        for project in self.app.model.projects():
            self.app.model.delete(project)

    def cleanup(self):
        for project in self.app.model.projects():
            self.app.model.delete(project)


app_classes = {'memory_cache': ToolboxTestApp,
               'file_cache': FileCacheTestApp,
               'couch': CouchTestApp}


def run_tests(app_cls,
              raise_on_error=False,
              cleanup=True,
              report_first=False,
              output=sys.stdout):

    results = {}

    # gather tests
    tests =  [ test for test in os.listdir(directory)
               if test.endswith('.txt') ]
    output.write("Tests:\n%s\n" % '\n'.join(tests))

    for test in tests:

        # create an app
        app = app_cls()

        # doctest arguments
        extraglobs = {'here': directory, 'app': app, 'urlescape': escape}
        doctest_args = dict(extraglobs=extraglobs, raise_on_error=raise_on_error)
        if report_first:
            doctest_args['optionflags'] = doctest.REPORT_ONLY_FIRST_FAILURE

        # run the test
        try:
            results[test] = doctest.testfile(test, **doctest_args)
        except doctest.DocTestFailure, failure:
            raise
        except doctest.UnexpectedException, failure:
            raise failure.exc_info[0], failure.exc_info[1], failure.exc_info[2]
        finally:
            if cleanup:
                app.cleanup()

    return results


def main(args=sys.argv[1:]):

    # parse command line args
    parser = OptionParser()
    parser.add_option('--no-cleanup', dest='cleanup',
                      default=True, action='store_false',
                      help="cleanup following the tests")
    parser.add_option('--raise', dest='raise_on_error',
                      default=False, action='store_true',
                      help="raise on first error")
    parser.add_option('--report-first', dest='report_first',
                      default=False, action='store_true',
                      help="report the first error only (all tests will still run)")
    parser.add_option('--model', dest='model', default='file_cache',
                      help="model to use")
    options, args = parser.parse_args(args)

    # get arguments to run_tests
    kw = dict([(i, getattr(options, i)) for i in ('raise_on_error', 'cleanup', 'report_first')])
    if options.model is not None:
        try:
            kw['app_cls'] = app_classes[options.model]
        except KeyError:
            parser.error("Model '%s' unknown (choose from: %s)" % (options.model, app_classes.keys()))

    # run the tests
    results = run_tests(**kw)
    if sum([i.failed for i in results.values()]):
        sys.exit(1) # error


if __name__ == '__main__':
    main()



"""
request handlers:
these are instantiated for every request, then called
"""

import cgi
import os
from datetime import datetime
from pkg_resources import resource_filename
from urllib import quote as _quote
from urlparse import urlparse
from util import strsplit
from util import JSONEncoder
from webob import Response, exc
from tempita import HTMLTemplate
from time import time

# this is necessary because WSGI stupidly follows the CGI convention wrt encoding slashes
# http://comments.gmane.org/gmane.comp.web.pylons.general/5922
encoded_slash = '%25%32%66'

def quote(s, safe='/'):
    if isinstance(s, unicode):
        s = s.encode('utf-8', 'ignore') # hope we're using utf-8!
    return _quote(s, safe).replace('/', encoded_slash)

try:
    import json
except ImportError:
    import simplejson as json

class HandlerMatchException(Exception):
    """the handler doesn't match the request"""

class Handler(object):
    """general purpose request handler (view)"""

    methods = set(['GET']) # methods to listen to
    handler_path = [] # path elements to match

    @classmethod
    def match(cls, app, request):

        # check the method
        if request.method not in cls.methods:
            return None

        # check the path
        if request.environ['path'] != cls.handler_path:
            return None

        # check the constructor
        try:
            return cls(app, request)
        except HandlerMatchException:
            return None
    
    def __init__(self, app, request):
        self.app = app
        self.request = request
        self.check_json() # is this a JSON request?

    def __call__(self):
        return getattr(self, self.request.method.title())()

    def link(self, path=None):
        """
        link relative to the site root
        """
        path_info = self.request.path_info
        segments = path_info.split('/')
        if segments[0]:
            segments.insert(0, '')

        if len(segments) <3:
            if not path or path == '/':
                return './'
            return path

        nlayers = len(segments[2:])
        string = '../' * nlayers

        if not path or path == '/':
            return string
        return string + path

    def redirect(self, location, query=None, anchor=None):
        return exc.HTTPSeeOther(location=self.app.baseurl + '/' + location
                                + (query and self.query_string(query) or '')
                                + (anchor and ('#' + anchor) or ''))

    def query_string(self, query):
        """
        generate a query string; query is a list of 2-tuples
        """
        return '?' + '&'.join(['%s=%s' % (i,j)
                               for i, j in query])

    # methods for JSON

    def check_json(self):
        """check to see if the request is for JSON"""
        self.json = self.request.GET.pop('format', '') == 'json'

    def post_data(self):
        """python dict from POST request"""
        if self.json:
            return json.loads(self.request.body)
        else:
            retval = self.request.POST.mixed()
            for key in retval:
                value = retval[key]
                if isinstance(value, basestring):
                    retval[key] = value.strip()
                else:
                    # TODO[?]: just throw away all empty values here
                    retval[key] = [i.strip() for i in value]
            return retval

    def get_json(self):
        """JSON to serialize if requested for GET"""
        raise NotImplementedError # abstract base class


class TempitaHandler(Handler):
    """handler for tempita templates"""

    template_dirs = [ resource_filename(__name__, 'templates') ]

    template_cache = {}

    css = ['css/html5boilerplate.css']

    less = ['css/style.less']

    js = ['js/jquery-1.6.min.js',
          'js/less-1.0.41.min.js',
          'js/jquery.timeago.js',
          'js/main.js']
    
    def __init__(self, app, request):
        Handler.__init__(self, app, request)

        # add application template_dir if specified
        if app.template_dir:
            self.template_dirs = self.template_dirs[:] + [app.template_dir]
            
        self.data = { 'request': request,
                      'css': self.css,
                      'item_name': self.app.item_name,
                      'item_plural': self.app.item_plural,
                      'less': self.less,
                      'js':  self.js,
                      'site_name': app.site_name,
                      'title': self.__class__.__name__,
                      'hasAbout': bool(app.about),
                      'urlescape': quote,
                      'link': self.link}

    def find_template(self, name):
        """find a template of a given name"""
        # the application caches a dict of the templates if app.reload is False
        if name in self.template_cache:
            return self.template_cache[name]
        
        for d in self.template_dirs:
            path = os.path.join(d, name)
            if os.path.exists(path):
                template = HTMLTemplate.from_filename(path)
                if not self.app.reload:
                    self.template_cache[name] = template
                return template

    def render(self, template, **data):
        template = self.find_template(template)
        if template is None:
            raise Exception("I can't find your template")
        return template.substitute(**data)

    def Get(self):
        # needs to have self.template set
        if self.json:
            return Response(content_type='application/json',
                            body=json.dumps(self.get_json(), cls=JSONEncoder))
        self.data['content'] = self.render(self.template, **self.data)
        return Response(content_type='text/html',
                        body=self.render('main.html', **self.data))

class ProjectsView(TempitaHandler):
    """abstract base class for views of projects"""

    js = TempitaHandler.js[:]
    js.extend(['js/jquery.tokeninput.js',
               'js/jquery.jeditable.js',
               'js/jquery.autolink.js',
               'js/project.js'])
               
    less = TempitaHandler.less[:]
    less.extend(['css/project.less'])

    css = TempitaHandler.css[:]
    css.extend(['css/token-input.css',
                'css/token-input-facebook.css'])

    def __init__(self, app, request):
        """project views specific init"""
        TempitaHandler.__init__(self, app, request)
        self.data['fields'] = self.app.model.fields()
        self.data['error'] = None
        if not self.json:
            self.data['format_date'] = self.format_date

    def get_json(self):
        """JSON to serialize if requested"""
        return self.data['projects']

    def sort(self, field):
        reverse = False
        if field.startswith('-'):
            field = field[1:]
            reverse = True
        if field == 'name':
            self.data['projects'].sort(key=lambda value: value[field].lower(), reverse=reverse)
        else:
            self.data['projects'].sort(key=lambda value: value[field], reverse=reverse)

    def format_date(self, timestamp):
        """return a string representation of a timestamp"""
        format_string = '%Y-%m-%dT%H:%M:%SZ'
        return datetime.utcfromtimestamp(timestamp).strftime(format_string)


class QueryView(ProjectsView):
    """general index view to query projects"""
    
    template = 'index.html'
    methods = set(['GET'])

    def __init__(self, app, request):
        ProjectsView.__init__(self, app, request)

        # pop non-query parameters;
        # sort is popped first so that it does go in the query
        sort_type = self.request.GET.pop('sort', None)
        query = self.request.GET.mixed()
        self.data['query'] = query
        search = query.pop('q', None)
        self.data['search'] = search

        # query for tools
        self.data['projects']= self.app.model.get(search, **query)

        # order the results
        self.data['sort_types'] = [('name', 'name'), ('-modified', 'last updated')]
        if search:
            self.data['sort_types'].insert(0, ('search', 'search rank'))
        if sort_type is None:
            if search:
                sort_type = 'search'
            else:
                # default
                sort_type = '-modified'
        self.data['sort_type'] = sort_type
        if sort_type != 'search':
            # preserve search order results 
            self.sort(sort_type)
            
        self.data['fields'] = self.app.model.fields()
        self.data['title'] = self.app.site_name


class ProjectView(ProjectsView):
    """view of a particular project"""

    template = 'index.html'
    methods=set(['GET', 'POST'])

    @classmethod
    def match(cls, app, request):

        # check the method
        if request.method not in cls.methods:
            return None

        # the path should match a project
        if not len(request.environ['path']) == 1:
            return None

        # get the project if it exists
        projectname = request.environ['path'][0].replace('%2f', '/')  # double de-escape slashes, see top of file
        try:
            # if its utf-8, we should try to keep it utf-8
            projectname = projectname.decode('utf-8')
        except UnicodeDecodeError:
            pass
        project = app.model.project(projectname)
        if not project:
            return None

        # check the constructor
        try:
            return cls(app, request, project)
        except HandlerMatchException:
            return None

    def __init__(self, app, request, project):
        ProjectsView.__init__(self, app, request)
        self.data['fields'] = self.app.model.fields()
        self.data['projects'] = [project]
        self.data['title'] = project['name']

    def get_json(self):
        return self.data['projects'][0]

    def Post(self):

        # data
        post_data = self.post_data()
        project = self.data['projects'][0]

        # insist that you have a name
        if 'name' in post_data and not post_data['name'].strip():
            self.data['title'] = 'Rename error'
            self.data['error'] = 'Cannot give a project an empty name'
            self.data['content'] = self.render(self.template, **self.data)
            return Response(content_type='text/html',
                            status=403,
                            body=self.render('main.html', **self.data))

        # don't allow overiding other projects with your fancy rename
        if 'name' in post_data and post_data['name'] != project['name']:
            if self.app.model.project(post_data['name']):
                self.data['title'] = '%s -> %s: Rename error' % (project['name'], post_data['name'])
                self.data['error'] = 'Cannot rename over existing project: <a href="%s">%s</a>' % (post_data['name'], post_data['name'] )
                self.data['content'] = self.render(self.template, **self.data)
                return Response(content_type='text/html',
                                status=403,
                                body=self.render('main.html', **self.data))

        # XXX for compatability with jeditable:
        id = post_data.pop('id', None)

        action = post_data.pop('action', None)
        old_name = project['name']
        if action == 'delete':
            for field in self.app.model.fields():
                if field in post_data and field in project:
                    values = post_data.pop(field)
                    if isinstance(values, basestring):
                        values = [values]
                    for value in values:
                        project[field].remove(value)
                    if not project[field]:
                        project.pop(field)
        else:
            for field in self.app.model.required:
                if field in post_data:
                    project[field] = post_data[field]
            for field in self.app.model.fields():
                if field in post_data:
                    value = post_data[field]
                    if isinstance(value, basestring):
                        value = strsplit(value)
                    if action == 'replace':
                        # replace the field from the POST request
                        project[field] = value
                    else:
                        # append the items....the default action
                        project.setdefault(field, []).extend(value)

        # rename handling
        if 'name' in post_data and post_data['name'] != old_name:
            self.app.model.delete(old_name)
            self.app.model.update(project)
            return self.redirect(quote(project['name']))

        self.app.model.update(project)

        # XXX for compatability with jeditable:
        if id is not None:
            return Response(content_type='text/plain',
                            body=cgi.escape(project['description']))

        # XXX should redirect instead
        return self.Get()


class FieldView(ProjectsView):
    """view of projects sorted by a field"""

    template = 'fields.html'
    methods=set(['GET', 'POST'])
    js = TempitaHandler.js[:] + ['js/field.js']

    @classmethod
    def match(cls, app, request):

        # check the method
        if request.method not in cls.methods:
            return None

        # the path should match a project
        if len(request.environ['path']) != 1:
            return None

        # ensure the field exists
        field = request.environ['path'][0]
        if field not in app.model.fields():
            return None

        # check the constructor
        try:
            return cls(app, request, field)
        except HandlerMatchException:
            return None

    def __init__(self, app, request, field):
        ProjectsView.__init__(self, app, request)
        projects = self.app.model.field_query(field)
        if projects is None:
            projects = {}
        self.data['field'] = field
        self.data['values'] = projects
        self.data['title'] = app.item_plural + ' by %s' % field
        if self.request.method == 'GET':
            # get project descriptions for tooltips
            descriptions = {}
            project_set = set()
            for values in projects.values():
                project_set.update(values)
            self.data['projects'] = dict([(name, self.app.model.project(name))
                                          for name in project_set])

    def Post(self):
        field = self.data['field']
        for key in self.request.POST.iterkeys():
            value = self.request.POST[key]
            self.app.model.rename_field_value(field, key, value)
        
        return self.redirect(field, anchor=value)
        
    def get_json(self):
        return self.data['values']

        
class CreateProjectView(TempitaHandler):
    """view to create a new project"""

    template = 'new.html'
    methods = set(['GET', 'POST'])
    handler_path = ['new']
    js = TempitaHandler.js[:]
    js.extend(['js/jquery.tokeninput.js',
               'js/queryString.js',
               'js/new.js'])
               
    less = TempitaHandler.less[:]
    less.extend(['css/new.less'])
    
    css = TempitaHandler.css[:]
    css.extend(['css/token-input.css',
                'css/token-input-facebook.css'])

    def __init__(self, app, request):
        TempitaHandler.__init__(self, app, request)
        self.data['title'] = 'Add a ' + app.item_name
        self.data['fields'] = self.app.model.fields()

    def check_name(self, name):
        """
        checks a project name for validity
        returns None on success or an error message if invalid
        """
        reserved = self.app.reserved.copy()
        if name in reserved or name in self.app.model.fields(): # check application-level reserved URLS
            return 'reserved'
        if self.app.model.project(name): # check projects for conflict
            return 'conflict'

    def Post(self):

        # get some data
        required = self.app.model.required
        post_data = self.post_data()

        # ensure the form isn't over 24 hours old
        day = 24*3600
        form_date = post_data.pop('form-render-date', -day)
        try:
            form_date = float(form_date)
        except ValueError:
            form_date = -day
        if abs(form_date - time()) > day:
            # if more than a day old, don't honor the request
            return Response(content_type='text/plain',
                            status=400,
                            body="Your form is over a day old or you don't have Javascript enabled")

        # build up a project dict
        project = dict([(i, post_data.get(i, '').strip())
                        for i in required])

        # check for errors
        errors = {}
        missing = set([i for i in required if not project[i]])
        if missing: # missing required fields
            errors['missing'] = missing
        # TODO check for duplicate project name
        # and other url namespace collisions
        name_conflict = self.check_name(project['name'])
        if name_conflict:
            errors[name_conflict] = [project['name']]
        if errors:
            error_list = []
            for key in errors:
                # flatten the error dict into a list
                error_list.extend([(key, i) for i in errors[key]])
            return self.redirect(self.request.path_info.strip('/'), error_list)

        # add fields to the project
        for field in self.app.model.fields():
            value = post_data.get(field, '').strip()
            values = strsplit(value)
            if not value:
                continue
            project[field] = values or value

        self.app.model.update(project)
        return self.redirect(quote(project['name']))


class DeleteProjectHandler(Handler):

    methods = set(['POST'])
    handler_path = ['delete']

    def Post(self):        
        post_data = self.post_data()
        project = post_data.get('project')
        if project:
            try:
                self.app.model.delete(project)
            except:
                pass # XXX better than internal server error

        # redirect to query view
        return self.redirect('')


class TagsView(TempitaHandler):
    """view most popular tags"""
    methods = set(['GET'])
    handler_path = ['tags']
    template = 'tags.html'

    def __init__(self, app, request):
        TempitaHandler.__init__(self, app, request)
        self.data['fields'] = self.app.model.fields()
        fields = self.request.GET.getall('field') or self.data['fields']
        query = self.request.GET.get('q', '')
        self.data['title'] = 'Tags'
        field_tags = dict((i, {}) for i in fields)
        omit = self.request.GET.getall('omit')
        ommitted = dict([(field, set()) for field in fields])
        for name in omit:
            project = self.app.model.project(name)
            if not project:
                continue
            for field in fields:
                ommitted[field].update(project.get(field, []))
            
        for project in self.app.model.get():
            if project in omit:
                continue
            # TODO: cache this for speed somehow
            # possibly at the model level
            for field in fields:
                for value in project.get(field, []):
                    if value in ommitted[field] or query not in value:
                        continue
                    count = field_tags[field].get(value, 0) + 1
                    field_tags[field][value] = count
        tags = []
        for field in field_tags:
            for value, count in field_tags[field].items():
                tags.append({'field': field, 'value': value, 'count': count, 'id': value, 'name': value})
        tags.sort(key=lambda x: x['count'], reverse=True)

        self.data['tags'] = tags

    def get_json(self):
        return self.data['tags']


class AboutView(TempitaHandler):
    """the obligatory about page"""
    methods = set(['GET'])
    handler_path = ['about']
    template = 'about.html'
    less = TempitaHandler.less[:] + ['css/about.less']
    def __init__(self, app, request):
        TempitaHandler.__init__(self, app, request)
        self.data['fields'] = self.app.model.fields()
        self.data['title'] = 'about:' + self.app.site_name
        self.data['about'] = self.app.about

class NotFound(TempitaHandler):
    def __init__(self, app, request):
        TempitaHandler.__init__(self, app, request)
        self.data['fields'] = self.app.model.fields()

    def __call__(self):
        self.data['content'] = '<h1 id="title">Not Found</h1>'
        return Response(content_type='text/html',
                        status=404,
                        body=self.render('main.html', **self.data))



import os
import shutil
import tempfile

from time import sleep
from whoosh import fields
from whoosh import index
from whoosh.query import And
from whoosh.query import Or
from whoosh.query import Term
from whoosh.qparser import QueryParser
from whoosh.index import LockError

class WhooshSearch(object):
    """full-text search"""

    def __init__(self, whoosh_index=None):
        """
        - whoosh_index : whoosh index directory
        """
        self.schema = fields.Schema(name=fields.ID(unique=True, stored=True),
                                    description=fields.TEXT)
        self.keywords = set([])
        self.tempdir = False
        if whoosh_index is None:
            whoosh_index = tempfile.mkdtemp()
            self.tempdir = True
        if not os.path.exists(whoosh_index):
            os.makedirs(whoosh_index)
        self.index = whoosh_index
        self.ix = index.create_in(self.index, self.schema)

    def update(self, name, description, **kw):
        """update a document"""

        # forgivingly get the writer
        timeout = 3. # seconds
        ctr = 0.
        incr = 0.2
        while ctr < timeout:
            try:
                writer = self.ix.writer()
                break
            except LockError:
                ctr += incr
                sleep(incr)
        else:
            raise

        # add keywords
        for key in kw:
            if key not in self.keywords:
                writer.add_field(key, fields.KEYWORD)
                self.keywords.add(key)
            if not isinstance(kw[key], basestring):
                kw[key] = ' '.join(kw[key])
            kw[key] = unicode(kw[key])

        # convert to unicode for whoosh
        # really whoosh should do this for us
        # and really python should be unicode-based :(
        name = unicode(name)
        description = unicode(description)

        writer.update_document(name=name, description=description, **kw)
        writer.commit()

    def delete(self, name):
        """delete a document of a given name"""
        writer = self.ix.writer()
        name = unicode(name)
        writer.delete_by_term('name', name)
        writer.commit()

    def __call__(self, query):
        """search"""
        query = unicode(query)
        query_parser = QueryParser("description", schema=self.ix.schema)
        myquery = query_parser.parse(query)

# Old code: too strict
#        extendedquery = Or([myquery] +
#                           [Term(field, query) for field in self.keywords])


        # New code: too permissive
#        extendedquery = [myquery]
        excluded = set(['AND', 'OR', 'NOT'])
        terms = [i for i in query.split() if i not in excluded]
#        for field in self.keywords:
#            extendedquery.extend([Term(field, term) for term in terms])
#        extendedquery = Or(extendedquery)

        # Code should look something like
        #Or([myquery] + [Or(
        # extendedquery = [myquery]
        extendedquery = And([Or([myquery] + [Term('description', term), Term('name', term)] +
                                [Term(field, term) for field in self.keywords]) for term in terms])

        # perform the search
        searcher = self.ix.searcher()
        return [i['name'] for i in searcher.search(extendedquery, limit=None)]
        
    def __del__(self):
        if self.tempdir:
            # delete the temporary directory, if present
            shutil.rmtree(self.index)



"""
request dispatcher WSGI app:
data persisting across requests should go here
"""

import os

from handlers import CreateProjectView
from handlers import DeleteProjectHandler
from handlers import FieldView
from handlers import ProjectView
from handlers import QueryView
from handlers import TagsView
from handlers import AboutView
from handlers import NotFound

from model import models
from util import strsplit
from webob import Request, Response, exc

here = os.path.dirname(os.path.abspath(__file__))

class Dispatcher(object):
    """toolbox WSGI app which dispatchers to associated handlers"""

    # class defaults
    defaults = { 'about': None, # file path to ReST about page
                 'model_type': 'memory_cache', # type of model to use
                 'handlers': None,
                 'reload': True, # reload templates
                 'reserved': None, # reserved URL namespaces
                 'template_dir': None, # directory for template overrides
                 'baseurl': '', # base URL for redirects

                 # branding variables
                 'site_name': 'toolbox', # name of the site
                 'item_name': 'tool', # name of a single item
                 'item_plural': None, # item_name's plural, or None for item_name + 's'
                 }

    def __init__(self, **kw):
        """
        **kw arguments used to override defaults
        additional **kw are passed to the model
        """

        # set instance parameters from kw and defaults
        for key in self.defaults:
            setattr(self, key, kw.pop(key, self.defaults[key]))
        if self.item_plural is None:
            self.item_plural = self.item_name + 's'

        # should templates be reloaded?
        if isinstance(self.reload, basestring):
            self.reload = self.reload.lower() == 'true'

        # model: backend storage and associated methods
        if 'fields' in kw and isinstance(kw['fields'], basestring):
            # split fields if given as a string
            kw['fields'] = strsplit(kw['fields'])
        if hasattr(self.model_type, '__call__'):
            model = self.model_type
        elif self.model_type in models:
            model = models[self.model_type]
        else:
            try:
                import pyloader
                model = pyloader.load(self.model_type)
            except:
                raise AssertionError("model_type '%s' not found in %s" % (self.model_type, models.keys()))
        self.model = model(**kw)

        # add an about view if file specified
        if self.about:
            about = file(self.about).read()
            import docutils.core
            about = docutils.core.publish_parts(about, writer_name='html')['body']
            self.about = about


        # request handlers in order they will be tried
        if self.handlers is None:
            self.handlers = [ TagsView,
                              CreateProjectView,
                              FieldView,
                              QueryView,
                              DeleteProjectHandler,
                              ProjectView]
            if self.about:
                self.handlers.append(AboutView)

        # extend reserved URLS from handlers
        if self.reserved is None:
            self.reserved = set(['css', 'js', 'img'])
            for handler in self.handlers:
                if handler.handler_path:
                    self.reserved.add(handler.handler_path[0])

    def __call__(self, environ, start_response):

        # get a request object
        request = Request(environ)

        # get the path 
        path = request.path_info.strip('/').split('/')
        if path == ['']:
            path = []
        request.environ['path'] = path

        # load any new data
        self.model.load()

        # match the request to a handler
        for h in self.handlers:
            handler = h.match(self, request)
            if handler is not None:
                break
        else:
            # our 404 handler
            handler = NotFound(self, request)

        # get response
        res = handler()
        return res(environ, start_response)



#!/usr/bin/env python

"""
WSGI -> HTTP server factories for toolbox
"""

import os
import sys

from dispatcher import Dispatcher
from paste.urlparser import StaticURLParser
from pkg_resources import resource_filename
from theslasher import TheSlasher

class PassthroughFileserver(object):
    """serve files if they exist"""

    def __init__(self, app, directory):
        self.app = app
        self.directory = directory
        self.fileserver = StaticURLParser(self.directory)

    def __call__(self, environ, start_response):
        path = environ['PATH_INFO'].strip('/')
        if path and os.path.exists(os.path.join(self.directory, path)) and '..' not in path:
            return self.fileserver(environ, start_response)
        return self.app(environ, start_response)


def paste_factory(global_conf=None, **app_conf):
    """create a webob view and wrap it in middleware"""

    keystr = 'toolbox.'
    static_directory = app_conf.pop('static',
                                    resource_filename(__name__, 'static'))
    args = dict([(key.split(keystr, 1)[-1], value)
                 for key, value in app_conf.items()
                 if key.startswith(keystr) ])
    app = TheSlasher(Dispatcher(**args)) # kill slashes
    return PassthroughFileserver(app, static_directory)

def wsgiref_factory(host='0.0.0.0', port=8080):
    """wsgiref factory; for testing only"""

    from wsgiref import simple_server
    app = Dispatcher()
    app = PassthroughFileserver(app, resource_filename(__name__, 'static'))
    server = simple_server.make_server(host=host, port=int(port), app=app)
    fqdn = '127.0.0.1' if host =='0.0.0.0' else host
    print "Serving toolbox at http://%s:%d/" % (fqdn, port)
    server.serve_forever()


# WSGI factories available
factories = {'paste': paste_factory,
             'wsgiref': wsgiref_factory}

def main(args=sys.argv[1:]):
    """CLI entry point"""

    # parse command line
    usage = '%prog [options]'
    import argparse
    parser = argparse.ArgumentParser(usage=usage, description=__doc__.strip())
    parser.add_argument('--factory', default='wsgiref',
                        choices=factories.keys(),
                        help="factory to use")
    parser.add_argument('--port', type=int, default=8080,
                        help="port to serve on")
    args = parser.parse_args()

    # serve toolbox
    factory = factories[args.__dict__.pop('factory')]
    factory_args = args.__dict__
    print "Serving using factory: %s" % getattr(factory, '__name__', str(factory))
    print "Factory arguments: %s" % factory_args
    factory(**factory_args)

if __name__ == '__main__':
    main()



"""
models for toolbox
"""

import couchdb
import os
import pyes
import sys
from copy import deepcopy
from search import WhooshSearch
from time import time
from util import str2filename

try:
    import json
except ImportError:
    import simplejson as json

# TODO: types of fields:
# - string: a single string: {'type': 'string', 'name': 'name', 'required': True}
# - field: a list of strings: {'type': 'field', 'name', 'usage'}
# - dict: a subclassifier: {'type': '???', 'name': 'url', 'required': True}
# - computed values, such as modified

class ProjectsModel(object):
    """
    abstract base class for toolbox tools
    """

    def __init__(self, fields=None, required=('name', 'description', 'url'),
                 whoosh_index=None):
        """
        - fields : list of fields to use, or None to calculate dynamically
        - required : required data (strings)
        - whoosh_index : directory to keep whoosh index in
        """
        self.required = set(required)

        # reserved fields
        self.reserved = self.required.copy()
        self.reserved.update(['modified']) # last modified, a computed value
        self.search = WhooshSearch(whoosh_index=whoosh_index)

        # classifier fields
        self._fields = fields
        self.field_set = set(fields or ())

    def update_search(self, project):
        """update the search index"""
        assert self.required.issubset(project.keys()) # XXX should go elsewhere
        fields = dict([(field, project[field])
                       for field in self.fields()
                       if field in project])

        # keys must be strings, not unicode, on some systems
        f = dict([(str(i), j) for i, j in fields.items()])

        self.search.update(name=project['name'], description=project['description'], **f)

    def fields(self):
        """what fields does the model support?"""
        if self._fields is not None:
            return self._fields
        return list(self.field_set)

    def projects(self):
        """list of all projects"""
        return [i['name'] for i in self.get()]

    def export(self, other):
        """export the current model to another model instance"""
        for project in self.get():
            other.update(project)

    def rename_field_value(self, field, from_value, to_value):
        projects = self.get(None, **{field: from_value})
        for project in projects:
            project[field].remove(from_value)
            project[field].append(to_value)
            self.update(project)

    ### implementor methods

    def update(self, project):
        """update a project"""
        raise NotImplementedError

    def get(self, search=None, **query):
        """
        get a list of projects matching a query
        the query should be key, value pairs to match;
        if the value is single, it should be a string;
        if the value is multiple, it should be a set which will be
        ANDed together
        """
        raise NotImplementedError

    def project(self, name):
        """get a project of a particular name, or None if there is none"""
        raise NotImplementedError

    def field_query(self, field):
        """get projects according to a particular field, or None"""
        raise NotImplementedError

    def delete(self, project):
        raise NotImplementedError


class MemoryCache(ProjectsModel):
    """
    sample implementation keeping everything in memory
    """

    def __init__(self, fields=None, whoosh_index=None):
        
        ProjectsModel.__init__(self, fields=fields, whoosh_index=whoosh_index)

        # indices
        self._projects = {}
        self.index = {}
        
        self.load()

    def update(self, project, load=False):
        
        if project['name'] in self._projects and project == self._projects[project['name']]:
            return # nothing to do
        if not load:
            project['modified'] = time()
        if self._fields is None:
            fields = [i for i in project if i not in self.reserved]
            self.field_set.update(fields)
        else:
            fields = self._fields
        for field in fields:
            for key, _set in self.index.get(field, {}).items():
                _set.discard(project['name'])
                if not _set:
                    self.index[field].pop(key)
            if field not in project:
                continue
            project[field] = list(set([i.strip() for i in project[field] if i.strip()]))
            index = self.index.setdefault(field, {})
            values = project[field]
            if isinstance(values, basestring):
                values = [values]
            for value in values:
                index.setdefault(value, set()).update([project['name']])
        self._projects[project['name']] = deepcopy(project)
        self.update_search(project)
        if not load:
            self.save(project)

    def get(self, search=None, **query):
        """
        - search: text search
        - query: fields to match
        """
        order = None
        if search:
            results = self.search(search)
            order = dict([(j,i) for i,j in enumerate(results)])
        else:
            results = self._projects.keys()
        results = set(results)
        for key, values in query.items():
            if isinstance(values, basestring):
                values = [values]
            for value in values:
                results.intersection_update(self.index.get(key, {}).get(value, set()))
        if order:
            # preserve search order
            results = sorted(list(results), key=lambda x: order[x])
        return [deepcopy(self._projects[project]) for project in results]


    def project(self, name):
        if name in self._projects:
            return deepcopy(self._projects[name])

    def field_query(self, field):
        if field in self.index:
            return deepcopy(self.index.get(field))

    def delete(self, project):
        """
        delete a project
        - project : name of the project
        """
        if project not in self._projects:
            return
        del self._projects[project]
        for field, classifiers in self.index.items():
            for key, values in classifiers.items():
                classifiers[key].discard(project)
                if not classifiers[key]:
                    del classifiers[key]
        self.search.delete(project)
        
    def load(self):
        """for subclasses; in memory, load nothing"""

    def save(self, project):
        """for subclasses; in memory, save nothing"""


class FileCache(MemoryCache):
    """save in JSON blob directory"""

    def __init__(self, directory, fields=None, whoosh_index=None):
        """
        - directory: directory of .json tool files
        """
        # JSON blob directory
        if not os.path.exists(directory):
            os.makedirs(directory)
        assert os.path.isdir(directory)
        self.directory = directory

        self.files = {}
        MemoryCache.__init__(self, fields=fields, whoosh_index=whoosh_index)

    def delete(self, project):
        MemoryCache.delete(self, project)
        os.remove(os.path.join(self.directory, self.files.pop(project)))

    def load(self):
        """load JSON from the directory"""
        for i in os.listdir(self.directory):
            if not i.endswith('.json'):
                continue
            filename = os.path.join(self.directory, i)
            try:
                project = json.loads(file(filename).read())
            except:
                print 'File: ' + i
                raise
            self.files[project['name']] = i
            self.update(project, load='modified' in project)

    def save(self, project):

        filename = self.files.get(project['name'])
        if not filename:
            filename = str2filename(project['name']) + '.json'
        filename = filename.encode('ascii', 'ignore')
        filename = os.path.join(self.directory, filename)
        try:
            f = file(filename, 'w')
        except Exception, e:
            print filename, repr(filename)
            raise
        f.write(json.dumps(project))
        f.close()


class ElasticSearchCache(MemoryCache):
    """
    store json in ElasticSearch
    """

    def __init__(self,
                 server="localhost:9200",
                 es_index="toolbox",
                 doc_type="projects",
                 fields=None,
                 whoosh_index=None):
        self.es_index = es_index
        self.doc_type = doc_type

        try:
            self.es = pyes.ES([server])
            self.es.create_index(self.es_index)
        except pyes.urllib3.connectionpool.MaxRetryError:
            raise Exception("Could not connect to ES instance")
        except pyes.exceptions.ElasticSearchException:
            # this just means the index already exists
            pass
        MemoryCache.__init__(self, fields=fields, whoosh_index=whoosh_index)

    def es_query(self, query):
        """make an ElasticSearch query and return the results"""
        search = pyes.Search(query)
        results = self.es.search(query=search,
                                 indexes=[self.es_index],
                                 doc_types=[self.doc_type],
                                 size=0)

        # the first query is just used to determine the size of the set
        if not 'hits' in results and not 'total' in results['hits']:
            raise Exception("bad ES response %s" % json.dumps(results))
        total = results['hits']['total']

        # repeat the query to retrieve the entire set
        results = self.es.search(query=search,
                                 indexes=[self.es_index],
                                 doc_types=[self.doc_type],
                                 size=total)

        if not 'hits' in results and not 'hits' in results['hits']:
            raise Exception("bad ES response %s" % json.dumps(results))

        return results

    def load(self):
        """load all json documents from ES:toolbox/projects"""
        query = pyes.MatchAllQuery()
        results = self.es_query(query)

        for hit in results['hits']['hits']:
            self.update(hit['_source'], True)

    def save(self, project):
        query = pyes.FieldQuery()
        query.add('name', project['name'])
        results = self.es_query(query)

        # If there is an existing records in ES with the same
        # project name, update that record.  Otherwise create a new record.
        id = None
        if results['hits']['hits']:
            id = results['hits']['hits'][0]['_id']

        self.es.index(project, self.es_index, self.doc_type, id)

    def delete(self, project):
        MemoryCache.delete(self, project)
        query = pyes.FieldQuery()
        query.add('name', project['name'])
        results = self.es_query(query)
        if results['hits']['hits']:
            id = results['hits']['hits'][0]['_id']
            self.es.delete(self.es_index, self.doc_type, id)


class CouchCache(MemoryCache):
    """
    store json files in couchdb
    """

    def __init__(self,
                 server="http://127.0.0.1:5984",
                 dbname="toolbox",
                 fields=None,
                 whoosh_index=None):

        # TODO: check if server is running
        couchserver = couchdb.Server(server)
        try:
            self.db = couchserver[dbname]
        except couchdb.ResourceNotFound: # XXX should not be a blanket except!
            self.db = couchserver.create(dbname)
        except:
            raise Exception("Could not connect to couch instance. Make sure that you have couch running at %s and that you have database create priveleges if '%s' does not exist" % (server, dbname))
        MemoryCache.__init__(self, fields=fields, whoosh_index=whoosh_index)

    def load(self):
        """load JSON objects from CouchDB docs"""
        for id in self.db:
            doc = self.db[id]
            try:
                project = doc['project']
            except KeyError:
                continue   # it's prob a design doc
            self.update(project, load=True)
            
    def save(self, project):
        name = project['name']
        try:
             updated = self.db[name]
        except:
             updated = {}
        updated['project'] = project
        self.db[name] = updated

    def delete(self, project):
        MemoryCache.delete(self, project)
        del self.db[project]

# directory of available models
models = {'memory_cache': MemoryCache,
          'file_cache': FileCache,
          'couch': CouchCache,
          'es': ElasticSearchCache}

def convert(args=sys.argv[1:]):
    """CLI front-end for model conversion"""
    from optparse import OptionParser
    usage = '%prog [global-options] from_model [options] to_model [options]'
    description = "export data from one model to another"
    parser = OptionParser(usage=usage, description=description)
    parser.disable_interspersed_args()
    parser.add_option('-l', '--list-models', dest='list_models',
                      action='store_true', default=False,
                      help="list available models")
    parser.add_option('-a', '--list-args', dest='list_args',
                      metavar='MODEL',
                      help="list arguments for a model")

    options, args = parser.parse_args(args)

    # process global options
    if options.list_models:
        for name in sorted(models.keys()):
            print name # could conceivably print docstring
        parser.exit()
    if options.list_args:
        if not options.list_args in models:
            parser.error("Model '%s' not found. (Choose from: %s)" % (options.list_args, models.keys()))
        ctor = models[options.list_args].__init__
        import inspect
        argspec = inspect.getargspec(ctor)
        defaults = [[i, None] for i in argspec.args[1:]] # ignore self
        for index, value in enumerate(reversed(argspec.defaults), 1):
            defaults[-index][-1] = value
        defaults = [[i,j] for i, j in defaults if i != 'fields']
        print '%s arguments:' % options.list_args
        for arg, value in defaults:
            print ' -%s %s' % (arg, value or '')
        parser.exit()

    # parse models and their ctor args
    sects = []
    _models = []
    for arg in args:
        if arg.startswith('-'):
            sects[-1].append(arg)
        else:
            _models.append(arg)
            sects.append([])

    # check models
    if len(_models) != 2:
        parser.error("Please provide two models. (You gave: %s)" % _models)
    if not set(_models).issubset(models):
        parser.error("Please use these models: %s (You gave: %s)" % (models, _models))

    sects = [ [i.lstrip('-') for i in sect ] for sect in sects ]

    # require an equals sign
    # XXX hacky but much easier to parse
    if [ True for sect in sects
         if [i for i in sect if '=' not in i] ]:
        parser.error("All arguments must be `key=value`")
    sects = [dict([i.split('=', 1) for i in sect]) for sect in sects]

    # instantiate models
    from_model = models[_models[0]](**sects[0])
    to_model = models[_models[1]](**sects[1])

    # convert the data
    from_model.export(to_model)

if __name__ == '__main__':
    convert()



from setuptools import setup

try:
    description = file('README.txt').read()
except IOError:
    description = ''

version = "0.4"

# dependencies
dependencies = [
    'WebOb',
    'tempita',
    'paste',
    'pastescript', # technically optional, but here for ease of install
    'whoosh >= 2.5',
    'couchdb',
    'docutils',
    'pyloader',
    'theslasher',
    'pyes == 0.15',
    ]

setup(name='toolk0s',
      version=version,
      description="a place to list links + tools",
      long_description=description,
      classifiers=[], # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      author='Jeff Hammel',
      author_email='jhammel@mozilla.com',
      url='http://k0s.org/hg/toolbox/',
      license="MPL",
      packages=['toolbox'],
      include_package_data=True,
      zip_safe=False,
      install_requires=dependencies,
      entry_points="""
      # -*- Entry points: -*-
      [console_scripts]
      toolbox-convert-model = toolbox.model:convert
      toolbox-serve = toolbox.factory:main

      [paste.app_factory]
      toolbox = toolbox.factory:paste_factory
      """,
      )



"""
utilities for toolbox
"""

try:
    import json
except ImportError:
    import simplejson as json

def strsplit(string):
    """sensibly split a comma-separated string"""
    string = string.strip()
    if not string:
        return []
    return [i.strip() for i in string.split(',')]

def strreplace(string, translation):
    """replace substrings from a translation matrix"""
    for key, value in translation.items():
        string = string.replace(key, value)
    return string

def str2filename(string):
    """converts a string to an acceptable filename"""
    matrix = {' ': '_',
              '>': '',
              '<': '',
              "'": '',
              '"': '',
              '&': '+',
              '\\': '',
              '\x00': '',
              '/': ''}
    return strreplace(string, matrix)


class JSONEncoder(json.JSONEncoder):
    """provide additional serialization for JSON"""

    def default(self, obj):
        if hasattr(obj, 'isoformat'):
            return obj.isoformat()
        if isinstance(obj, set):
            return list(obj)

        return json.JSONEncoder.default(self, obj)

if __name__ == '__main__':
    # test the encoder
    testjson = {}

    # test date encoding
    from datetime import datetime
    testjson['date'] = datetime.now()

    # test set encoding
    testjson['set'] = set([1,2,3,2])

    print json.dumps(testjson, cls=JSONEncoder)
